///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// This file is part of ModelBlocks. Copyright 2009, ModelBlocks developers. //
//                                                                           //
//    ModelBlocks is free software: you can redistribute it and/or modify    //
//    it under the terms of the GNU General Public License as published by   //
//    the Free Software Foundation, either version 3 of the License, or      //
//    (at your option) any later version.                                    //
//                                                                           //
//    ModelBlocks is distributed in the hope that it will be useful,         //
//    but WITHOUT ANY WARRANTY; without even the implied warranty of         //
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          //
//    GNU General Public License for more details.                           //
//                                                                           //
//    You should have received a copy of the GNU General Public License      //
//    along with ModelBlocks.  If not, see <http://www.gnu.org/licenses/>.   //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "nl-cpt.h"
#include "nl-dtree.h"
using namespace std;
//
////////////////////////////////////////////////////////////////////////////////

//// P: part of speech category...
//// it seems that I don't need this since I do have Pos in my HModel
//DiscreteDomain<short> domainP;
//typedef DiscreteDomainRV<short,domainP> P;


//the following arguments do not work well, but can be improved
//-v genmodel/indmodel.pronountrain genmodel/genmodel.pronountrain genmodel/nummodel.pronountrain genmodel/synmodel.pronountrain genmodel/opmodel.train genmodel/posmodel.pronountrain /home/dingcheng/Documents/NLPWorkspace/wsjparse/genmodel/posWordDepmodel.train genmodel/wordDepModel.pro genmodel/wordModel.pro genmodel/349699.pronountrain


////////////////////////////////////////////////////////////////////////////////
//
//  Models
//
////////////////////////////////////////////////////////////////////////////////

//// Preterminal (POS) given constituent category models...
/// I know why G can be used here directly since G is a class from HHMMLangModel-?.h
/// when compiling HHMMLangModel-?.cpp which include both HHMMLangModel-?.h and TextObsModel.h,
/// Therefore, the compiler can find both classes in both files.
/// This is a point very different from Java in which, each file involves only one main class.
/// At this point, I confused java files and cpp files and thus failed to understand why G can be used
/// directly here. In fact, it is like call a java class within the same package.
/// Dingcheng's study notes Sept. 16, 09.


//// Preterminal (POS) given constituent category models...
//typedef HidVarCPT2DModel<Pos,C,LogProb> PgivCModel;

const int MAXELEMENT = 12;
//string pronArray[MAXELEMENT] = {"she", "her", "hers","he","him","his","they","their","theirs","them","it","its","this","that","these","those","whose","whom"};
string pronArray[MAXELEMENT] = { "she", "her", "hers", "he", "him", "his", "they", "their", "theirs", "them", "it", "its" };
List<W> pronList;
static SimpleHash<W, int> pronSiHash;

void fillPronHash() {
	pronList = List<W> ();
	for (int ii = 0; ii < MAXELEMENT; ii++) {
		W w(pronArray[ii].c_str());
		//pronSiHash.set(w)=1;
		pronList.add() = w;
	}
}

//// Generative model of word given tag...
//class WModel {
//private:
//	TrainableDTree2DModel<Pos,W,LogProb> modPgivWdt;
//	RandAccCPT2DModel<Pos,W,LogProb> modPgivWs;
//	RandAccCPT1DModel<Pos,LogProb> modP;
//	RandAccCPT1DModel<W,LogProb> modW;
//
//public:
//	//LogProb getProb ( const W& w, const HidVarCPT1DModel<P,LogProb>::IterVal& p ) const {
//	LogProb getProb ( const W& w, const Pos::ArrayIterator<LogProb>& p ) const {
//		assert(modP.getProb(p)!=LogProb());
//		LogProb pr = ( (  modW.contains(w) ? modPgivWs.getProb(p,w) : modPgivWdt.getProb(p,w) )
//				* LogProb(-1000) / modP.getProb(p) );
//		return pr;
//	}
//	void writeFields ( FILE* pf, string sPref ) { modPgivWdt.writeFields(pf,sPref); }
//	friend pair<StringInput,WModel*> operator>> ( StringInput si, WModel& m ) { return pair<StringInput,WModel*>(si,&m); }
//	friend StringInput operator>> ( pair<StringInput,WModel*> delimbuff, const char* psD ) {
//		StringInput si;
//		//the following lines need changed.
//		return ( (si=delimbuff.first>>"W "   >>delimbuff.second->modW      >>psD)!=NULL ||
//				(si=delimbuff.first>>"Pw "  >>delimbuff.second->modPgivWs >>psD)!=NULL ||
//				(si=delimbuff.first>>"PwDT ">>delimbuff.second->modPgivWdt>>psD)!=NULL ||
//				(si=delimbuff.first>>"P "   >>delimbuff.second->modP      >>psD)!=NULL ) ? si : StringInput(NULL);
//	}
//};

DiscreteDomain<short> domainIndex;
typedef DiscreteDomainRV<short, domainIndex> EntIndex;

//////////////////////////////////////// "Wrapper" models for individual RVs...
typedef RandAccCPT2DModel<NewGen, W, LogProb> NewGenGivWModel;
typedef RandAccCPT2DModel<NewNum, W, LogProb> NewNumGivWModel;
typedef RandAccCPT2DModel<NewSyn, W, LogProb> NewSynGivWModel;
typedef RandAccCPT2DModel<Pos, W, LogProb> PosGivWModel;
typedef RandAccCPT2DModel<W, EntIndex, LogProb> IndexGivWModel;

class OModel {
private:
	//WModel     modWgivP;
	NewGenGivWModel mGen_W;
	NewNumGivWModel mNum_W;
	NewSynGivWModel mSyn_W;
	PosGivWModel mPos_W;
	IndexGivWModel mEntity_W;
	TrainableDTree2DModel<Pos, W, LogProb> modPgivWdt;
	RandAccCPT1DModel<NewGen, LogProb> mGender;
	RandAccCPT1DModel<NewNum, LogProb> mNumber;
	RandAccCPT1DModel<NewSyn, LogProb> mSyntax;
	RandAccCPT1DModel<Pos, LogProb> mPos;
	RandAccCPT1DModel<W, LogProb> mW;
public:
	class DistribModeledWgivG {
	private:
		W wW;
		EntIndex indexW;
		LogProb pr;
	public:

		DistribModeledWgivG& set(const W& w, const EntIndex & index, const OModel& m) {
			wW = w;
			indexW = index;
			return *this;
		}
		//Prob&   setProb ( const Pos & pos )       { return hcpCache.set(pos); }
		//LogProb getProb ( const Pos & pos ) const { return LogProb(hcpCache.get(pos)); }
		LogProb getProb(const OModel& m, const S & s) const {
			return m.calcProb(*this, s);
		}
		W getW() const {
			return wW;
		}
		EntIndex getIndex() const {
			return indexW;
		}
	};

	typedef DistribModeledWgivG RandVarType;
	LogProb calcProb(const OModel::RandVarType& o, const S& s) const {
		//fillPronList();
		W w = o.getW();
		EntIndex indexW = o.getIndex();
		//		if(w==W("and")){
		//			cerr<<"start the debugging and : "<<endl;
		//		}
		//cerr<<"\nindexW: "<<indexW<<endl;
		//Pos::ArrayIterator<LogProb> pos;
		LogProb pr;
		//cerr<<"mEntity_w.getProb(indexW,w): "<<mEntity_W.getProb(w, indexW)<<" LogProb: "<<LogProb()<<" s.first: "<<s.first<<" w: "<<w<<" indexW: "<<indexW<<endl;
		if (w == W("eos")) {
			pr = mGen_W.getProb(NewGen("-"), w);
			pr *= mNum_W.getProb(NewNum("-"), w);
			pr *= mSyn_W.getProb(NewSyn("-"), w);
			pr *= mPos_W.getProb(Pos("-"), w);
			pr *= mW.getProb(w);
			pr /= mGender.getProb(NewGen("-"));
			pr /= mNumber.getProb(NewNum("-"));
			pr /= mSyntax.getProb(NewSyn("-"));
			pr /= mPos.getProb(Pos("-"));
		} else if ((mEntity_W.getProb(w, indexW) == 0 && s.first == OP(COPY)) || (mEntity_W.getProb(w, indexW) == LogProb() && s.first != OP(COPY))) {
			pr = LogProb();
			//cerr<<" pr when it is not an entity but op is suppsed to be entity:  "<<pr<<endl;
		} else if ((mEntity_W.getProb(w, indexW) == 0 && s.first != OP(COPY)) || (mEntity_W.getProb(w, indexW) == LogProb() && s.first == OP(COPY))) {
			//			if(w==W("stand")){
			//				cerr<<"start the debugging: "<<endl;
			//			}
			if (mW.contains(w)) {
				if (pronList.contains(w) && s.first != OP(OLD)) {
					pr = LogProb();
				} else if (s.first == OP(OLD) || s.first == OP(NEW)) {
					int size = atoi(Size(s.second.first.first.second).getString().c_str());
					NewGen newGen = Gen(s.second.first.second).get(size - 1);
					NewNum newNum = Num(s.second.first.third).get(size - 1);
					NewSyn newSyn = Syn(s.second.first.fourth).get(size - 1);
					cerr << "size: " << size << " newGen: " << newGen << " newNum: " << newNum << " newSyn: " << newSyn << endl;
					pr = mGen_W.getProb(newGen, w);
					cerr << " gen: " << s.second.first.second.first.getString().substr(s.second.first.second.first.getString().find_last_of("_") + 1) << " word:  " << w << " Pr: " << pr << endl;
					pr *= mNum_W.getProb(newNum, w);
					cerr << " num: " << s.second.first.third.first.getString().substr(s.second.first.third.first.getString().find_last_of("_") + 1) << " word:  " << w << " pr: " << pr << endl;
					pr *= mSyn_W.getProb(newSyn, w);
					cerr << "syn: " << s.second.first.fourth.first.getString().substr(s.second.first.fourth.first.getString().find_last_of("_") + 1) << " word:  " << w << " pr: " << pr << endl;
					//pr*=mPos_W.getProb(Pos(s.second.second.getString().substr(s.second.second.getString().find_last_of("_")+1).c_str()),w);
					pr *= mPos_W.getProb(s.second.second, w);
					cerr << "pos: " << s.second.second.getString().substr(s.second.second.getString().find_last_of("_") + 1) << " word:  " << w << " pr: " << pr << endl;
					pr *= mW.getProb(w);
					cerr << "mW: " << " word:  " << w << " pr: " << pr << endl;
					pr /= mGender.getProb(newGen);
					cerr << "mGender: " << pr << endl;
					pr /= mNumber.getProb(newNum);
					cerr << "mNumber: " << pr << endl;
					pr /= mSyntax.getProb(newSyn);
					cerr << "mSyntax: " << pr << endl;
					pr /= mPos.getProb(Pos(s.second.second.getString().substr(s.second.second.getString().find_last_of("_") + 1).c_str()));
					cerr << "mPos: " << pr << endl;
					cout << " pr when there may be an entity an entity but op is suppsed to be entity: " << OP(s.first) << " pr: " << pr << endl;
				} else {
					pr = mPos_W.getProb(s.second.second, w);
					cerr << "mPos: " << s.second.second << " w: " << w << " pr: " << pr << endl;
					pr *= mW.getProb(w);
					cerr << "mW: " << w << " pr " << pr << endl;
					pr /= mPos.getProb(s.second.second);
					cerr << "mPos: " << w << " pr " << pr << endl;
				}
			} else {
				RandAccCPT1DModel<Pos, LogProb> modP;
				Pos::ArrayIterator<LogProb> pos;
				assert(modP.getProb(pos)!=LogProb());
				if ((s.first == OP(OLD) || s.first == OP(NEW))) {

					//				pr=p=mPgivWdt.getProb(w,pos);
					//				pr*=p=LogProb(-1000);
					//				pr/=mPos.getProb(pos);
					W wUnknown = W("unkword");
					//the following way get probability with d-tree
					if (Pos(pos) == Pos(s.second.second)) {
						pr = modPgivWdt.getProb(pos, w);
						//pr=modWgivP.getProb(w,pos);
					} else {
						pr = LogProb();
					}
					int size = atoi(Size(s.second.first.first.second).getString().c_str());
					NewGen newGen = Gen(s.second.first.second).get(size - 1);
					NewNum newNum = Num(s.second.first.third).get(size - 1);
					NewSyn newSyn = Syn(s.second.first.fourth).get(size - 1);
					cerr << "size: " << size << " newGen: " << newGen << " newNum: " << newNum << " newSyn: " << newSyn << endl;
					//pr=mPos_W.getProb(s.second.second,wUnknown);
					cerr << "pos: " << s.second.first << " word:  " << w << " pr: " << pr << endl;
					pr *= mGender.getProb(newGen);
					cerr << "mGender: " << pr << endl;
					pr *= mNumber.getProb(newNum);
					cerr << "mNumber: " << pr << endl;
					pr *= mSyntax.getProb(newSyn);
					pr *= mW.getProb(wUnknown);

					pr /= mPos.getProb(s.second.second);
				} else {
					//this is a question, pos here is not the same as s.second.second. How can I
					//transfer this back to s?
					Pos::ArrayIterator<LogProb> pos;
					//				pr=p=mPgivWdt.getProb(w,pos);
					//				pr*=p=LogProb(-1000);
					//				pr/=mPos.getProb(pos);
					W wUnknown = W("unkword");
					//call d-tree to get unknown word probability. In fact, modWgivP will make redundant judgment.
					if (Pos(pos) == Pos(s.second.second)) {
						pr = modPgivWdt.getProb(pos, w);
						//pr=modWgivP.getProb(w,pos);
					} else {
						pr = LogProb();
					}
					//pr=mPos_W.getProb(s.second.second,wUnknown);
					//	cerr<<"pos: "<<s.second.first<<" word:  "<<w<<" pr: "<<pr<<endl;
					pr *= mW.getProb(wUnknown);
					pr /= mPos.getProb(s.second.second);
				} //cout<<" unknown words, but it is not: "<<endl;
			}
		} else if ((mEntity_W.getProb(w, indexW).toProb() == 0.5)) {
			if (s.first == OP(COPY)) {
				pr = mPos_W.getProb(s.second.second, w);
				cerr << "mPos: " << s.second.second << " w: " << w << " pr: " << pr << endl;
				pr *= mW.getProb(w);
				cerr << "mW: " << w << " pr " << pr << endl;
				pr /= mPos.getProb(s.second.second);
				cerr << "mPos: " << w << " pr " << pr << endl;
			} else {
				pr = LogProb();
			}
		}
		return pr;
	}

	LogProb getProb(const OModel::RandVarType& o, const S & s) const {
		return o.getProb(*this, s);
	}

	friend pair<StringInput, OModel*> operator>>(StringInput si, OModel& m) {
		return pair<StringInput, OModel*> (si, &m);
	}
	friend StringInput operator>>(pair<StringInput, OModel*> delimbuff, const char* psD) {
		StringInput si;
		return ((si = delimbuff.first >> "Gen_W " >> delimbuff.second->mGen_W >> psD) != NULL || (si = delimbuff.first >> "Num_W " >> delimbuff.second->mNum_W >> psD) != NULL ||
		//(si=delimbuff.first>>"Syn_W " >>delimbuff.second->mSyn_W>>psD)!=NULL  ||
				(si = delimbuff.first >> "Ne_W " >> delimbuff.second->mSyn_W >> psD) != NULL || (si = delimbuff.first >> "Pos_W " >> delimbuff.second->mPos_W >> psD) != NULL || (si = delimbuff.first >> "Gender " >> delimbuff.second->mGender >> psD) != NULL || (si = delimbuff.first >> "Number " >> delimbuff.second->mNumber >> psD) != NULL ||
		//(si=delimbuff.first>>"Syntax " >>delimbuff.second->mSyntax>>psD)!=NULL ||
				(si = delimbuff.first >> "Name " >> delimbuff.second->mSyntax >> psD) != NULL || (si = delimbuff.first >> "Word " >> delimbuff.second->mW >> psD) != NULL || (si = delimbuff.first >> "Entity " >> delimbuff.second->mEntity_W >> psD) != NULL || (si = delimbuff.first >> "PwDT " >> delimbuff.second->modPgivWdt >> psD) != NULL || (si = delimbuff.first >> "Pos " >> delimbuff.second->mPos >> psD) != NULL ? si : StringInput(NULL));
	}
	//void writeFields ( FILE* pf, string sPref ) { modWgivP.writeFields(pf,sPref); }
};
